from typing import Dict, List, Tuple, Type, TYPE_CHECKING
if TYPE_CHECKING:
    from ..client import CQiClient
    from ..status import StatusOk
    from .corpora import Corpus
from .resource import Collection, Model


class Attribute(Model):
    @property
    def api_name(self) -> str:
        return self.attrs['api_name']

    @property
    def name(self) -> str:
        return self.attrs['name']

    @property
    def size(self) -> int:
        return self.attrs['size']

    def drop(self) -> 'StatusOk':
        ''' unload attribute from memory '''
        return self.client.api.cl_drop_attribute(self.api_name)


class AttributeCollection(Collection):
    model: Type[Attribute] = Attribute

    def __init__(self, client: 'CQiClient' = None, corpus: 'Corpus' = None):
        super().__init__(client=client)
        self.corpus: 'Corpus' = corpus

    def _get(self, attribute_name: str) -> Dict:
        api_name: str = f'{self.corpus.api_name}.{attribute_name}'
        return {
            'api_name': api_name,
            'name': attribute_name,
            'size': self.client.api.cl_attribute_size(api_name)
        }

    def get(self, attribute_name: str) -> Attribute:
        return self.prepare_model(self._get(attribute_name))

    def list(self) -> List[Attribute]:
        raise NotImplementedError


class AlignmentAttribute(Attribute):
    def cpos_by_id(self, id: int) -> Tuple[int, int, int, int]:
        ''' returns (src_start, src_end, target_start, target_end) '''
        return self.client.api.cl_alg2cpos(self.api_name, id)

    def ids_by_cpos(self, cpos_list: List[int]) -> List[int]:
        ''' returns -1 for every corpus position not inside an alignment '''
        return self.client.api.cl_cpos2alg(self.api_name, cpos_list)


class AlignmentAttributeCollection(AttributeCollection):
    model: Type[AlignmentAttribute] = AlignmentAttribute

    def list(self) -> List[AlignmentAttribute]:
        return [
            self.get(x) for x in
            self.client.api.corpus_alignment_attributes(self.corpus.api_name)
        ]


class PositionalAttribute(Attribute):
    @property
    def lexicon_size(self) -> int:
        return self.attrs['lexicon_size']

    def cpos_by_id(self, id: int) -> List[int]:
        ''' returns all corpus positions where the given token occurs '''
        return self.client.api.cl_id2cpos(self.api_name, id)

    def cpos_by_ids(self, id_list: List[int]) -> List[int]:
        '''
        returns all corpus positions where one of the tokens in <id_list>
        occurs; the returned list is sorted as a whole, not per token id
        '''
        return self.client.api.cl_idlist2cpos(self.api_name, id_list)

    def freqs_by_ids(self, id_list: List[int]) -> List[int]:
        ''' returns 0 for every ID in <id_list> that is out of range '''
        return self.client.api.cl_id2freq(self.api_name, id_list)

    def ids_by_cpos(self, cpos_list: List[int]) -> List[int]:
        '''
        returns -1 for every corpus position in <cpos_list> that is out of
        range
        '''
        return self.client.api.cl_cpos2id(self.api_name, cpos_list)

    def ids_by_regex(self, regex: str) -> List[int]:
        '''
        returns lexicon IDs of all tokens that match <regex>; the returned
        list may be empty (size 0);
        '''
        return self.client.api.cl_regex2id(self.api_name, regex)

    def ids_by_values(self, value_list: List[str]) -> List[int]:
        '''
        returns -1 for every string in <value_list> that is not found in the
        lexicon
        '''
        return self.client.api.cl_str2id(self.api_name, value_list)

    def values_by_cpos(self, cpos_list: List[int]) -> List[str]:
        '''
        returns "" for every corpus position in <cpos_list> that is out of
        range
        '''
        return self.client.api.cl_cpos2str(self.api_name, cpos_list)

    def values_by_ids(self, id_list: List[int]) -> List[str]:
        ''' returns "" for every ID in <id_list> that is out of range '''
        return self.client.api.cl_id2str(self.api_name, id_list)


class PositionalAttributeCollection(AttributeCollection):
    model: Type[PositionalAttribute] = PositionalAttribute

    def _get(self, positional_attribute_name: str) -> Dict:
        attrs = super()._get(positional_attribute_name)
        attrs['lexicon_size'] = self.client.api.cl_lexicon_size(attrs['api_name'])
        return attrs

    def list(self) -> List[PositionalAttribute]:
        return [
            self.get(x) for x in
            self.client.api.corpus_positional_attributes(self.corpus.api_name)
        ]


class StructuralAttribute(Attribute):
    @property
    def has_values(self) -> bool:
        return self.attrs['has_values']

    def cpos_by_id(self, id: int) -> Tuple[int, int]:
        '''
        returns start and end corpus positions of structure region with id
        <id>
        '''
        return self.client.api.cl_struc2cpos(self.api_name, id)

    def ids_by_cpos(self, cpos_list: List[int]) -> List[int]:
        '''
        returns -1 for every corpus position not inside a structure region
        '''
        return self.client.api.cl_cpos2struc(self.api_name, cpos_list)

    def lbound_by_cpos(self, cpos_list: List[int]) -> List[int]:
        '''
        returns left boundary of s-attribute region enclosing cpos, -1 if not
        in region
        '''
        return self.client.api.cl_cpos2lbound(self.api_name, cpos_list)

    def rbound_by_cpos(self, cpos_list: List[int]) -> List[int]:
        '''
        returns right boundary of s-attribute region enclosing cpos, -1 if not
        in region
        '''
        return self.client.api.cl_cpos2rbound(self.api_name, cpos_list)

    def values_by_ids(self, id_list: List[int]) -> List[str]:
        '''
        returns annotated string values of structure regions in <id_list>; ""
        if out of range

        check has_values property first
        '''
        return self.client.api.cl_struc2str(self.api_name, id_list)


class StructuralAttributeCollection(AttributeCollection):
    model: Type[StructuralAttribute] = StructuralAttribute

    def _get(self, structural_attribute_name: str) -> Dict:
        attrs = super()._get(structural_attribute_name)
        attrs['has_values'] = self.client.api.corpus_structural_attribute_has_values(attrs['api_name'])
        return attrs

    def list(self, filters: Dict = {}) -> List[StructuralAttribute]:
        structural_attributes = [
            self.get(x) for x in
            self.client.api.corpus_structural_attributes(self.corpus.api_name)
        ]
        for k, v in filters.items():
            if k == 'has_values':
                structural_attributes = [
                    x for x in structural_attributes
                    if x.has_values == v
                ]
            elif k == 'part_of':
                structural_attributes = [
                    x for x in structural_attributes
                    if x.name.startswith(f'{v.name}_')
                ]
        return structural_attributes



from typing import Callable, Dict, List, Tuple
import socket
import struct
import time
from . import specification
from .. import errors
from .. import status


class APIClient:
    '''
    A low-level client for the IMS Open Corpus Workbench (CWB) corpus query
    interface (CQi) API.

    Example:
    >>> import cqi
    >>> client = cqi.APIClient('127.0.0.1')
    >>> client.ctrl_connect('user', 'password')
    258  # CQI_STATUS_CONNECT_OK
    >>> client.ctrl_ping()
    260  # CQI_STATUS_PING_OK
    >>> client.ctrl_bye()
    259  # CQI_STATUS_BYE_OK

    Attributes:
    host (str): URL to the CQP server. For example,
        ``cqpserver.localhost`` or ``127.0.0.1``.
    port (int): Port the CQP server listens on. Default: ``4877``
    socket (socket.socket): Socket for communicating with a CQP server.
    timeout (float): Default timeout for API calls, in seconds.
        Default: ``60.0``
    version (str): The version of the CQi protocol to use. Default: ``0.1``
    '''

    def __init__(
        self,
        host: str,
        port: int = 4877,
        timeout: float = 60.0,
        version: str = '0.1'
    ):
        self.host: str = host
        self.port: int = port
        self.socket: socket.socket = socket.socket()
        self.timeout: float = timeout
        self.version: str = version

    def ctrl_connect(
        self,
        username: str,
        password: str
    ) -> status.StatusConnectOk:
        self.socket.connect((self.host, self.port))
        self.__send_WORD(specification.CTRL_CONNECT)
        self.__send_STRING(username)
        self.__send_STRING(password)
        return self.__recv_response()

    def ctrl_bye(self) -> status.StatusByeOk:
        self.__send_WORD(specification.CTRL_BYE)
        response: status.StatusByeOk = self.__recv_response()
        self.socket.close()
        return response

    def ctrl_user_abort(self):
        self.__send_WORD(specification.CTRL_USER_ABORT)

    def ctrl_ping(self) -> status.StatusPingOk:
        self.__send_WORD(specification.CTRL_PING)
        return self.__recv_response()

    def ctrl_last_general_error(self) -> str:
        ''' 
        Full-text error message for the last general error reported by the CQi
        server
        '''
        self.__send_WORD(specification.CTRL_LAST_GENERAL_ERROR)
        return self.__recv_response()

    def ask_feature_cqi_1_0(self) -> bool:
        self.__send_WORD(specification.ASK_FEATURE_CQI_1_0)
        return self.__recv_response()

    def ask_feature_cl_2_3(self) -> bool:
        self.__send_WORD(specification.ASK_FEATURE_CL_2_3)
        return self.__recv_response()

    def ask_feature_cqp_2_3(self) -> bool:
        self.__send_WORD(specification.ASK_FEATURE_CL_2_3)
        return self.__recv_response()

    def corpus_list_corpora(self) -> List[str]:
        self.__send_WORD(specification.CORPUS_LIST_CORPORA)
        return self.__recv_response()

    def corpus_charset(self, corpus: str) -> str:
        self.__send_WORD(specification.CORPUS_CHARSET)
        self.__send_STRING(corpus)
        return self.__recv_response()

    def corpus_properties(self, corpus: str) -> List[str]:
        self.__send_WORD(specification.CORPUS_PROPERTIES)
        self.__send_STRING(corpus)
        return self.__recv_response()

    def corpus_positional_attributes(self, corpus: str) -> List[str]:
        self.__send_WORD(specification.CORPUS_POSITIONAL_ATTRIBUTES)
        self.__send_STRING(corpus)
        return self.__recv_response()

    def corpus_structural_attributes(self, corpus: str) -> List[str]:
        self.__send_WORD(specification.CORPUS_STRUCTURAL_ATTRIBUTES)
        self.__send_STRING(corpus)
        return self.__recv_response()

    def corpus_structural_attribute_has_values(self, attribute: str) -> bool:
        self.__send_WORD(specification.CORPUS_STRUCTURAL_ATTRIBUTE_HAS_VALUES)
        self.__send_STRING(attribute)
        return self.__recv_response()

    def corpus_alignment_attributes(self, corpus: str) -> List[str]:
        self.__send_WORD(specification.CORPUS_ALIGNMENT_ATTRIBUTES)
        self.__send_STRING(corpus)
        return self.__recv_response()

    def corpus_full_name(self, corpus: str) -> str:
        ''' the full name of <corpus> as specified in its registry entry '''
        self.__send_WORD(specification.CORPUS_FULL_NAME)
        self.__send_STRING(corpus)
        return self.__recv_response()

    def corpus_info(self, corpus: str) -> List[str]:
        ''' 
        returns the contents of the .info file of <corpus> as a list of lines
        '''
        self.__send_WORD(specification.CORPUS_INFO)
        self.__send_STRING(corpus)
        return self.__recv_response()

    def corpus_drop_corpus(self, corpus: str) -> status.StatusOk:
        ''' try to unload a corpus and all its attributes from memory '''
        self.__send_WORD(specification.CORPUS_DROP_CORPUS)
        self.__send_STRING(corpus)
        return self.__recv_response()

    def cl_attribute_size(self, attribute: str) -> int:
        ''' 
        returns the size of <attribute>:
        - number of tokens        (positional)
        - number of regions       (structural)
        - number of alignments    (alignment)
        '''
        self.__send_WORD(specification.CL_ATTRIBUTE_SIZE)
        self.__send_STRING(attribute)
        return self.__recv_response()

    def cl_lexicon_size(self, attribute: str) -> int:
        '''
        returns the number of entries in the lexicon of a positional
        attribute;

        valid lexicon IDs range from 0 .. (lexicon_size - 1)
        '''
        self.__send_WORD(specification.CL_LEXICON_SIZE)
        self.__send_STRING(attribute)
        return self.__recv_response()

    def cl_drop_attribute(self, attribute: str) -> status.StatusOk:
        '''
        unload attribute from memory

        Note: Not implemented on the server side:
              https://sourceforge.net/p/cwb/code/HEAD/tree/cwb/trunk/CQi/cqpserver.c#l356
        '''
        raise NotImplementedError
        self.__send_WORD(specification.CL_DROP_ATTRIBUTE)
        self.__send_STRING(attribute)
        return self.__recv_response()

    '''
    ' NOTE: simple (scalar) mappings are applied to lists (the returned list
    '       has exactly the same length as the list passed as an argument)
    '''

    def cl_str2id(self, attribute: str, strings: List[str]) -> List[int]:
        '''
        returns -1 for every string in <strings> that is not found in the
        lexicon
        '''
        self.__send_WORD(specification.CL_STR2ID)
        self.__send_STRING(attribute)
        self.__send_STRING_LIST(strings)
        return self.__recv_response()

    def cl_id2str(self, attribute: str, id: List[int]) -> List[str]:
        ''' returns "" for every ID in <id> that is out of range '''
        self.__send_WORD(specification.CL_ID2STR)
        self.__send_STRING(attribute)
        self.__send_INT_LIST(id)
        return self.__recv_response()

    def cl_id2freq(self, attribute: str, id: List[int]) -> List[int]:
        ''' returns 0 for every ID in <id> that is out of range '''
        self.__send_WORD(specification.CL_ID2FREQ)
        self.__send_STRING(attribute)
        self.__send_INT_LIST(id)
        return self.__recv_response()

    def cl_cpos2id(self, attribute: str, cpos: List[int]) -> List[int]:
        ''' 
        returns -1 for every corpus position in <cpos> that is out of range
        '''
        self.__send_WORD(specification.CL_CPOS2ID)
        self.__send_STRING(attribute)
        self.__send_INT_LIST(cpos)
        return self.__recv_response()

    def cl_cpos2str(self, attribute: str, cpos: List[int]) -> List[str]:
        '''
        returns "" for every corpus position in <cpos> that is out of range
        '''
        self.__send_WORD(specification.CL_CPOS2STR)
        self.__send_STRING(attribute)
        self.__send_INT_LIST(cpos)
        return self.__recv_response()

    def cl_cpos2struc(self, attribute: str, cpos: List[int]) -> List[int]:
        '''
        returns -1 for every corpus position not inside a structure region
        '''
        self.__send_WORD(specification.CL_CPOS2STRUC)
        self.__send_STRING(attribute)
        self.__send_INT_LIST(cpos)
        return self.__recv_response()

    '''
    ' NOTE: temporary addition for the Euralex2000 tutorial, but should
    '       probably be included in CQi specs
    '''

    def cl_cpos2lbound(self, attribute: str, cpos: List[int]) -> List[int]:
        '''
        returns left boundary of s-attribute region enclosing cpos, -1 if not
        in region
        '''
        self.__send_WORD(specification.CL_CPOS2LBOUND)
        self.__send_STRING(attribute)
        self.__send_INT_LIST(cpos)
        return self.__recv_response()

    def cl_cpos2rbound(self, attribute: str, cpos: List[int]) -> List[int]:
        '''
        returns right boundary of s-attribute region enclosing cpos, -1 if not
        in region
        '''
        self.__send_WORD(specification.CL_CPOS2RBOUND)
        self.__send_STRING(attribute)
        self.__send_INT_LIST(cpos)
        return self.__recv_response()

    def cl_cpos2alg(self, attribute: str, cpos: List[int]) -> List[int]:
        ''' returns -1 for every corpus position not inside an alignment '''
        self.__send_WORD(specification.CL_CPOS2ALG)
        self.__send_STRING(attribute)
        self.__send_INT_LIST(cpos)
        return self.__recv_response()

    def cl_struc2str(self, attribute: str, strucs: List[int]) -> List[str]:
        '''
        returns annotated string values of structure regions in <strucs>;
        "" if out of range

        check corpus_structural_attribute_has_values(<attribute>) first
        '''
        self.__send_WORD(specification.CL_STRUC2STR)
        self.__send_STRING(attribute)
        self.__send_INT_LIST(strucs)
        return self.__recv_response()

    '''
    ' NOTE: the following mappings take a single argument and return multiple
    '       values, including lists of arbitrary size
    '''

    def cl_id2cpos(self, attribute: str, id: int) -> List[int]:
        ''' returns all corpus positions where the given token occurs '''
        self.__send_WORD(specification.CL_ID2CPOS)
        self.__send_STRING(attribute)
        self.__send_INT(id)
        return self.__recv_response()

    def cl_idlist2cpos(self, attribute: str, id_list: List[int]) -> List[int]:
        '''
        returns all corpus positions where one of the tokens in <id_list>
        occurs; the returned list is sorted as a whole, not per token id
        '''
        self.__send_WORD(specification.CL_IDLIST2CPOS)
        self.__send_STRING(attribute)
        self.__send_INT_LIST(id_list)
        return self.__recv_response()

    def cl_regex2id(self, attribute: str, regex: str) -> List[int]:
        '''
        returns lexicon IDs of all tokens that match <regex>; the returned
        list may be empty (size 0);
        '''
        self.__send_WORD(specification.CL_REGEX2ID)
        self.__send_STRING(attribute)
        self.__send_STRING(regex)
        return self.__recv_response()

    def cl_struc2cpos(self, attribute: str, struc: int) -> Tuple[int, int]:
        '''
        returns start and end corpus positions of structure region <struc>
        '''
        self.__send_WORD(specification.CL_STRUC2CPOS)
        self.__send_STRING(attribute)
        self.__send_INT(struc)
        return self.__recv_response()

    def cl_alg2cpos(
        self,
        attribute:
        str,
        alg: int
    ) -> Tuple[int, int, int, int]:
        ''' returns (src_start, src_end, target_start, target_end) '''
        self.__send_WORD(specification.CL_ALG2CPOS)
        self.__send_STRING(attribute)
        self.__send_INT(alg)
        return self.__recv_response()

    def cqp_query(self,
        mother_corpus: str,
        subcorpus_name: str,
        query: str
    ) -> status.StatusOk:
        ''' <query> must include the ';' character terminating the query. '''
        self.__send_WORD(specification.CQP_QUERY)
        self.__send_STRING(mother_corpus)
        self.__send_STRING(subcorpus_name)
        self.__send_STRING(query)
        return self.__recv_response()

    def cqp_list_subcorpora(self, corpus: str) -> List[str]:
        self.__send_WORD(specification.CQP_LIST_SUBCORPORA)
        self.__send_STRING(corpus)
        return self.__recv_response()

    def cqp_subcorpus_size(self, subcorpus: str) -> int:
        self.__send_WORD(specification.CQP_SUBCORPUS_SIZE)
        self.__send_STRING(subcorpus)
        return self.__recv_response()

    def cqp_subcorpus_has_field(self, subcorpus: str, field: int) -> bool:
        self.__send_WORD(specification.CQP_SUBCORPUS_HAS_FIELD)
        self.__send_STRING(subcorpus)
        self.__send_BYTE(field)
        return self.__recv_response()

    def cqp_dump_subcorpus(
        self,
        subcorpus: str,
        field: int,
        first: int,
        last: int
    ) -> List[int]:
        '''
        Dump the values of <field> for match ranges <first> .. <last> in
        <subcorpus>. <field> is one of the CQI_CONST_FIELD_* constants.
        '''
        self.__send_WORD(specification.CQP_DUMP_SUBCORPUS)
        self.__send_STRING(subcorpus)
        self.__send_BYTE(field)
        self.__send_INT(first)
        self.__send_INT(last)
        return self.__recv_response()

    def cqp_drop_subcorpus(self, subcorpus: str) -> status.StatusOk:
        ''' delete a subcorpus from memory '''
        self.__send_WORD(specification.CQP_DROP_SUBCORPUS)
        self.__send_STRING(subcorpus)
        return self.__recv_response()

    '''
    ' NOTE: The following two functions are temporarily included for the
    '       Euralex 2000 tutorial demo
    '''

    def cqp_fdist_1(
        self,
        subcorpus: str,
        cutoff: int,
        field: int,
        attribute: str
    ) -> List[int]:
        '''
        frequency distribution of single tokens

        returns <n> (id, frequency) pairs flattened into a list of size 2*<n>

        field is one of
        - CQI_CONST_FIELD_MATCH
        - CQI_CONST_FIELD_TARGET
        - CQI_CONST_FIELD_KEYWORD

        NB: pairs are sorted by frequency desc.
        '''
        self.__send_WORD(specification.CQP_FDIST_1)
        self.__send_STRING(subcorpus)
        self.__send_INT(cutoff)
        self.__send_BYTE(field)
        self.__send_STRING(attribute)
        return self.__recv_response()

    def cqp_fdist_2(
        self,
        subcorpus: str,
        cutoff: int,
        field1: int,
        attribute1: str,
        field2: int,
        attribute2: str
    ) -> List[int]:
        '''
        frequency distribution of pairs of tokens

        returns <n> (id1, id2, frequency) pairs flattened into a list of size
        3*<n>

        NB: triples are sorted by frequency desc.
        '''
        self.__send_WORD(specification.CQP_FDIST_2)
        self.__send_STRING(subcorpus)
        self.__send_INT(cutoff)
        self.__send_BYTE(field1)
        self.__send_STRING(attribute1)
        self.__send_BYTE(field2)
        self.__send_STRING(attribute2)
        return self.__recv_response()

    def __recv_response(self):
        byte_data: int = self.__recv_WORD()
        response_type: int = byte_data >> 8
        if response_type == specification.DATA:
            try:
                return self.__recv_DATA_method_lookup[byte_data](self)
            except KeyError:
                raise errors.CQiException(f'Unknown data type: {byte_data}')
        elif response_type == specification.STATUS:
            try:
                return status.lookup[byte_data]()
            except KeyError:
                raise errors.CQiException(f'Unknown status code: {byte_data}')
        elif response_type == specification.ERROR:
            try:
                raise errors.lookup[byte_data]()
            except KeyError:
                raise errors.CQiException(f'Unknown error code: {byte_data}')
        else:
            raise errors.CQiException(
                f'Unknown response type: {response_type}'
            )

    def __recv(self, bufsize: int):
        # This method should not be necessary but
        # self.socket.recv(bufsize, socket.MSG_WAITALL)
        # does not work for some reason
        start_time = time.time()
        while time.time() - start_time < self.timeout:
            # Check if the server already sent over the desired number of bytes
            if len(self.socket.recv(bufsize, socket.MSG_PEEK)) == bufsize:
                return self.socket.recv(bufsize)
            # Wait a bit before checking again
            time.sleep(0.05)
        raise TimeoutError()

    def __recv_DATA_BYTE(self) -> int:
        byte_data: bytes = self.__recv(1)
        return struct.unpack('!B', byte_data)[0]

    def __recv_DATA_BOOL(self) -> bool:
        byte_data: bytes = self.__recv(1)
        return struct.unpack('!?', byte_data)[0]

    def __recv_DATA_INT(self) -> int:
        byte_data: bytes = self.__recv(4)
        return struct.unpack('!i', byte_data)[0]

    def __recv_DATA_STRING(self) -> str:
        n: int = self.__recv_WORD()
        byte_data: bytes = self.__recv(n)
        # return byte_data.decode()
        return struct.unpack(f'!{n}s', byte_data)[0].decode()

    def __recv_DATA_BYTE_LIST(self) -> List[int]:
        data: List[int] = []
        n: int = self.__recv_DATA_INT()
        while n > 0:
            data.append(self.__recv_DATA_BYTE())
            n -= 1
        return data

    def __recv_DATA_BOOL_LIST(self) -> List[bool]:
        data: List[bool] = []
        n: int = self.__recv_DATA_INT()
        while n > 0:
            data.append(self.__recv_DATA_BOOL())
            n -= 1
        return data

    def __recv_DATA_INT_LIST(self) -> List[int]:
        data: List[int] = []
        n: int = self.__recv_DATA_INT()
        while n > 0:
            data.append(self.__recv_DATA_INT())
            n -= 1
        return data

    def __recv_DATA_STRING_LIST(self) -> List[str]:
        data: List[str] = []
        n: int = self.__recv_DATA_INT()
        while n > 0:
            data.append(self.__recv_DATA_STRING())
            n -= 1
        return data

    def __recv_DATA_INT_INT(self) -> Tuple[int, int]:
        return (self.__recv_DATA_INT(), self.__recv_DATA_INT())

    def __recv_DATA_INT_INT_INT_INT(self) -> Tuple[int, int, int, int]:
        return (
            self.__recv_DATA_INT(),
            self.__recv_DATA_INT(),
            self.__recv_DATA_INT(),
            self.__recv_DATA_INT()
        )

    def __recv_DATA_INT_TABLE(self) -> List[List[int]]:
        rows: int = self.__recv_DATA_INT()
        columns: int = self.__recv_DATA_INT()
        data: List[List[int]] = []
        for i in range(0, rows):
            row: List[int] = []
            for j in range(0, columns):
                row.append(self.__recv_DATA_INT())
            data.append(row)
        return data

    __recv_DATA_method_lookup: Dict[int, Callable] = {
        769: __recv_DATA_BYTE,
        770: __recv_DATA_BOOL,
        771: __recv_DATA_INT,
        772: __recv_DATA_STRING,
        773: __recv_DATA_BYTE_LIST,
        774: __recv_DATA_BOOL_LIST,
        775: __recv_DATA_INT_LIST,
        776: __recv_DATA_STRING_LIST,
        777: __recv_DATA_INT_INT,
        778: __recv_DATA_INT_INT_INT_INT,
        779: __recv_DATA_INT_TABLE
    }

    def __recv_WORD(self) -> int:
        byte_data: bytes = self.__recv(2)
        return struct.unpack('!H', byte_data)[0]

    def __send_BYTE(self, byte_data: int):
        data: bytes = struct.pack('!B', byte_data)
        self.socket.sendall(data)

    def __send_BOOL(self, bool_data: bool):
        data: bytes = struct.pack('!?', bool_data)
        self.socket.sendall(data)

    def __send_INT(self, int_data: int):
        data: bytes = struct.pack('!i', int_data)
        self.socket.sendall(data)

    def __send_STRING(self, string_data: str):
        encoded_string_data: bytes = string_data.encode('utf-8')
        n: int = len(encoded_string_data)
        data: bytes = struct.pack(f'!H{n}s', n, encoded_string_data)
        self.socket.sendall(data)

    def __send_INT_LIST(self, int_list_data: List[int]):
        n: int = len(int_list_data)
        self.__send_INT(n)
        for int_data in int_list_data:
            self.__send_INT(int_data)

    def __send_STRING_LIST(self, string_list_data: List[str]):
        n: int = len(string_list_data)
        self.__send_INT(n)
        for string_data in string_list_data:
            self.__send_STRING(string_data)

    def __send_WORD(self, word_data: int):
        data: bytes = struct.pack('!H', word_data)
        self.socket.sendall(data)



from typing import Dict, List, Type, TYPE_CHECKING
if TYPE_CHECKING:
    from ..status import StatusOk
from .attributes import (
    AlignmentAttributeCollection,
    PositionalAttributeCollection,
    StructuralAttributeCollection
)
from .resource import Collection, Model
from .subcorpora import SubcorpusCollection


class Corpus(Model):
    @property
    def api_name(self) -> str:
        return self.attrs['api_name']

    @property
    def name(self) -> str:
        return self.attrs['name']

    @property
    def size(self) -> int:
        return self.attrs['size']

    @property
    def charset(self) -> str:
        return self.attrs['charset']

    @property
    def properties(self) -> List[str]:
        return self.attrs['properties']

    @property
    def alignment_attributes(self) -> AlignmentAttributeCollection:
        return AlignmentAttributeCollection(client=self.client, corpus=self)

    @property
    def positional_attributes(self) -> PositionalAttributeCollection:
        return PositionalAttributeCollection(client=self.client, corpus=self)

    @property
    def structural_attributes(self) -> StructuralAttributeCollection:
        return StructuralAttributeCollection(client=self.client, corpus=self)

    @property
    def subcorpora(self) -> SubcorpusCollection:
        return SubcorpusCollection(client=self.client, corpus=self)

    def drop(self) -> 'StatusOk':
        ''' try to unload a corpus and all its attributes from memory '''
        return self.client.api.corpus_drop_corpus(self.api_name)

    def query(self, subcorpus_name: str, query: str) -> 'StatusOk':
        ''' <query> must include the ';' character terminating the query. '''
        return self.client.api.cqp_query(self.api_name, subcorpus_name, query)


class CorpusCollection(Collection):
    model: Type[Corpus] = Corpus

    def _get(self, corpus_name: str) -> Dict:
        api_name: str = corpus_name
        p_attr_names: List[str] = self.client.api.corpus_positional_attributes(api_name)
        corpus_size: int = (
            0 if len(p_attr_names) == 0 else
            self.client.api.cl_attribute_size(f'{api_name}.{p_attr_names[0]}')
        )
        return {
            'api_name': api_name,
            'charset': self.client.api.corpus_charset(api_name),
            # 'full_name': self.client.api.corpus_full_name(api_name),
            # 'info': self.client.api.corpus_info(api_name),
            'name': corpus_name,
            'properties': self.client.api.corpus_properties(api_name),
            'size': corpus_size
        }

    def get(self, corpus_name: str) -> Corpus:
        return self.prepare_model(self._get(corpus_name))

    def list(self) -> List[Corpus]:
        return [self.get(x) for x in self.client.api.corpus_list_corpora()]



from typing import Dict, Optional, Type
from .api import specification


class CQiStatus:
    ''' A base class from which all other status inherit. '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: Optional[int] = None


class StatusOk(CQiStatus):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.STATUS_OK


class StatusConnectOk(CQiStatus):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.STATUS_CONNECT_OK


class StatusByeOk(CQiStatus):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.STATUS_BYE_OK


class StatusPingOk(CQiStatus):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.STATUS_PING_OK


lookup: Dict[int, Type[CQiStatus]] = {
    specification.STATUS_OK: StatusOk,
    specification.STATUS_CONNECT_OK: StatusConnectOk,
    specification.STATUS_BYE_OK: StatusByeOk,
    specification.STATUS_PING_OK: StatusPingOk
}



""" 1. padding """
PAD = 0x00


""" 2. CQi responses """
""" 2.1 CQI_STATUS_* """
STATUS = 0x01
STATUS_OK = 0x0101
STATUS_CONNECT_OK = 0x0102
STATUS_BYE_OK = 0x0103
STATUS_PING_OK = 0x0104

""" 2.2 CQI_ERROR_*  """
ERROR = 0x02
ERROR_GENERAL_ERROR = 0x0201
ERROR_CONNECT_REFUSED = 0x0202
ERROR_USER_ABORT = 0x0203
ERROR_SYNTAX_ERROR = 0x0204
# includes corpus/attribute/subcorpus specifier syntax

""" 2.3 CQI_DATA_* """
DATA = 0x03
DATA_BYTE = 0x0301
DATA_BOOL = 0x0302
DATA_INT = 0x0303
DATA_STRING = 0x0304
DATA_BYTE_LIST = 0x0305
DATA_BOOL_LIST = 0x0306
DATA_INT_LIST = 0x0307
DATA_STRING_LIST = 0x0308
DATA_INT_INT = 0x0309
DATA_INT_INT_INT_INT = 0x030A
DATA_INT_TABLE = 0x030B

""" 2.4 CQI_CL_ERROR_* """
"""
" NOTE: some CL error codes are not represented in the CQi specs
"       - usually because they're not used in the CL any more
"       - CDA_ENOSTRING is not considered an error (returns -1)
"       - CDA_EARGS: dynamic attribute calls not yet supported
"""
CL_ERROR = 0x04
CL_ERROR_NO_SUCH_ATTRIBUTE = 0x0401
# returned if CQi server couldn't open attribute
CL_ERROR_WRONG_ATTRIBUTE_TYPE = 0x0402
# CDA_EATTTYPE
CL_ERROR_OUT_OF_RANGE = 0x0403
# CDA_EIDORNG, CDA_EIDXORNG, CDA_EPOSORNG
CL_ERROR_REGEX = 0x0404
# CDA_EPATTERN (not used), CDA_EBADREGEX
CL_ERROR_CORPUS_ACCESS = 0x0405
# CDA_ENODATA
CL_ERROR_OUT_OF_MEMORY = 0x0406
# CDA_ENOMEM
# this means the CQi server has run out of memory;
# try discarding some other corpora and/or subcorpora
CL_ERROR_INTERNAL = 0x0407
# CDA_EOTHER, CDA_ENYI
# this is the classical 'please contact technical support' error

""" 2.5 CQI_CQP_ERROR_* """
CQP_ERROR = 0x05
# CQP error messages yet to be defined
CQP_ERROR_GENERAL = 0x0501
CQP_ERROR_NO_SUCH_CORPUS = 0x0502
CQP_ERROR_INVALID_FIELD = 0x0503
CQP_ERROR_OUT_OF_RANGE = 0x0504
# various cases where a number is out of range


""" 3. CQi commands """
""" 3.1 CQI_CTRL_* """
CTRL = 0x11
CTRL_CONNECT = 0x1101
# INPUT: (STRING username, STRING password)
# OUTPUT: CQI_STATUS_CONNECT_OK, CQI_ERROR_CONNECT_REFUSED
CTRL_BYE = 0x1102
# INPUT: ()
# OUTPUT: CQI_STATUS_BYE_OK
CTRL_USER_ABORT = 0x1103
# INPUT: ()
# OUTPUT:
CTRL_PING = 0x1104
# INPUT: ()
# OUTPUT: CQI_STATUS_PING_OK
CTRL_LAST_GENERAL_ERROR = 0x1105
# INPUT: ()
# OUTPUT: CQI_DATA_STRING
# full-text error message for the last general error reported by the CQi server

""" 3.2 CQI_ASK_FEATURE_* """
ASK_FEATURE = 0x12
ASK_FEATURE_CQI_1_0 = 0x1201
# INPUT: ()
# OUTPUT: CQI_DATA_BOOL
ASK_FEATURE_CL_2_3 = 0x1202
# INPUT: ()
# OUTPUT: CQI_DATA_BOOL
ASK_FEATURE_CQP_2_3 = 0x1203
# INPUT: ()
# OUTPUT: CQI_DATA_BOOL

""" 3.3 CQI_CORPUS_* """
CORPUS = 0x13
CORPUS_LIST_CORPORA = 0x1301
# INPUT: ()
# OUTPUT: CQI_DATA_STRING_LIST
CORPUS_CHARSET = 0x1303
# INPUT: (STRING corpus)
# OUTPUT: CQI_DATA_STRING
CORPUS_PROPERTIES = 0x1304
# INPUT: (STRING corpus)
# OUTPUT: CQI_DATA_STRING_LIST
CORPUS_POSITIONAL_ATTRIBUTES = 0x1305
# INPUT: (STRING corpus)
# OUTPUT: CQI_DATA_STRING_LIST
CORPUS_STRUCTURAL_ATTRIBUTES = 0x1306
# INPUT: (STRING corpus)
# OUTPUT: CQI_DATA_STRING_LIST
CORPUS_STRUCTURAL_ATTRIBUTE_HAS_VALUES = 0x1307
# INPUT: (STRING attribute)
# OUTPUT: CQI_DATA_BOOL
CORPUS_ALIGNMENT_ATTRIBUTES = 0x1308
# INPUT: (STRING corpus)
# OUTPUT: CQI_DATA_STRING_LIST
CORPUS_FULL_NAME = 0x1309
# INPUT: (STRING corpus)
# OUTPUT: CQI_DATA_STRING
# the full name of <corpus> as specified in its registry entry
CORPUS_INFO = 0x130A
# INPUT: (STRING corpus)
# OUTPUT: CQI_DATA_STRING_LIST
# returns the contents of the .info file of <corpus> as a list of lines
CORPUS_DROP_CORPUS = 0x130B
# INPUT: (STRING corpus)
# OUTPUT: CQI_STATUS_OK
# try to unload a corpus and all its attributes from memory

""" 3.4 CQI_CL_* """
CL = 0x14
# low-level corpus access (CL functions)
CL_ATTRIBUTE_SIZE = 0x1401
# INPUT: (STRING attribute)
# OUTPUT: CQI_DATA_INT
# returns the size of <attribute>:
# - number of tokens (positional)
# - number of regions (structural)
# - number of alignments (alignment)
CL_LEXICON_SIZE = 0x1402
# INPUT: (STRING attribute)
# OUTPUT: CQI_DATA_INT
# returns the number of entries in the lexicon of a positional attribute;
# valid lexicon IDs range from 0 .. (lexicon_size - 1)
CL_DROP_ATTRIBUTE = 0x1403
# INPUT: (STRING attribute)
# OUTPUT: CQI_STATUS_OK
# unload attribute from memory
"""
" NOTE: simple (scalar) mappings are applied to lists (the returned list has
"       exactly the same length as the list passed as an argument)
"""
CL_STR2ID = 0x1404
# INPUT: (STRING attribute, STRING_LIST strings)
# OUTPUT: CQI_DATA_INT_LIST
# returns -1 for every string in <strings> that is not found in the lexicon
CL_ID2STR = 0x1405
# INPUT: (STRING attribute, INT_LIST id)
# OUTPUT: CQI_DATA_STRING_LIST
# returns "" for every ID in <id> that is out of range
CL_ID2FREQ = 0x1406
# INPUT: (STRING attribute, INT_LIST id)
# OUTPUT: CQI_DATA_INT_LIST
# returns 0 for every ID in <id> that is out of range
CL_CPOS2ID = 0x1407
# INPUT: (STRING attribute, INT_LIST cpos)
# OUTPUT: CQI_DATA_INT_LIST
# returns -1 for every corpus position in <cpos> that is out of range
CL_CPOS2STR = 0x1408
# INPUT: (STRING attribute, INT_LIST cpos)
# OUTPUT: CQI_DATA_STRING_LIST
# returns "" for every corpus position in <cpos> that is out of range
CL_CPOS2STRUC = 0x1409
# INPUT: (STRING attribute, INT_LIST cpos)
# OUTPUT: CQI_DATA_INT_LIST
# returns -1 for every corpus position not inside a structure region
"""
" NOTE: temporary addition for the Euralex2000 tutorial, but should probably be
"       included in CQi specs
"""
CL_CPOS2LBOUND = 0x1420
# INPUT: (STRING attribute, INT_LIST cpos)
# OUTPUT: CQI_DATA_INT_LIST
# returns left boundary of s-attribute region enclosing cpos, -1 if not in
# region
CL_CPOS2RBOUND = 0x1421
# INPUT: (STRING attribute, INT_LIST cpos)
# OUTPUT: CQI_DATA_INT_LIST
# returns right boundary of s-attribute region enclosing cpos, -1 if not in
# region
CL_CPOS2ALG = 0x140A
# INPUT: (STRING attribute, INT_LIST cpos)
# OUTPUT: CQI_DATA_INT_LIST
# returns -1 for every corpus position not inside an alignment
CL_STRUC2STR = 0x140B
# INPUT: (STRING attribute, INT_LIST strucs)
# OUTPUT: CQI_DATA_STRING_LIST
# returns annotated string values of structure regions in <strucs>; "" if out
# of range
# check CQI_CORPUS_STRUCTURAL_ATTRIBUTE_HAS_VALUES(<attribute>) first
"""
" NOTE: the following mappings take a single argument and return multiple
"       values, including lists of arbitrary size
"""
CL_ID2CPOS = 0x140C
# INPUT: (STRING attribute, INT id)
# OUTPUT: CQI_DATA_INT_LIST
# returns all corpus positions where the given token occurs
CL_IDLIST2CPOS = 0x140D
# INPUT: (STRING attribute, INT_LIST id_list)
# OUTPUT: CQI_DATA_INT_LIST
# returns all corpus positions where one of the tokens in <id_list>
# occurs; the returned list is sorted as a whole, not per token id
CL_REGEX2ID = 0x140E
# INPUT: (STRING attribute, STRING regex)
# OUTPUT: CQI_DATA_INT_LIST
# returns lexicon IDs of all tokens that match <regex>; the returned
# list may be empty (size 0);
CL_STRUC2CPOS = 0x140F
# INPUT: (STRING attribute, INT struc)
# OUTPUT: CQI_DATA_INT_INT
# returns start and end corpus positions of structure region <struc>
CL_ALG2CPOS = 0x1410
# INPUT: (STRING attribute, INT alg)
# OUTPUT: CQI_DATA_INT_INT_INT_INT
# returns (src_start, src_end, target_start, target_end)

""" 3.5 CQI_CQP_* """
CQP = 0x15
CQP_QUERY = 0x1501
# INPUT: (STRING mother_corpus, STRING subcorpus_name, STRING query)
# OUTPUT: CQI_STATUS_OK
# <query> must include the ';' character terminating the query.
CQP_LIST_SUBCORPORA = 0x1502
# INPUT: (STRING corpus)
# OUTPUT: CQI_DATA_STRING_LIST
CQP_SUBCORPUS_SIZE = 0x1503
# INPUT: (STRING subcorpus)
# OUTPUT: CQI_DATA_INT
CQP_SUBCORPUS_HAS_FIELD = 0x1504
# INPUT: (STRING subcorpus, BYTE field)
# OUTPUT: CQI_DATA_BOOL
CQP_DUMP_SUBCORPUS = 0x1505
# INPUT: (STRING subcorpus, BYTE field, INT first, INT last)
# OUTPUT: CQI_DATA_INT_LIST
# Dump the values of <field> for match ranges <first> .. <last> in <subcorpus>.
# <field> is one of the CQI_CONST_FIELD_* constants.
CQP_DROP_SUBCORPUS = 0x1509
# INPUT: (STRING subcorpus)
# OUTPUT: CQI_STATUS_OK
# delete a subcorpus from memory
"""
" NOTE: The following two functions are temporarily included for the Euralex
"       2000 tutorial demo
"""
""" NOTE: frequency distribution of single tokens """
CQP_FDIST_1 = 0x1510
# INPUT: (STRING subcorpus, INT cutoff, BYTE field, STRING attribute)
# OUTPUT: CQI_DATA_INT_LIST
# returns <n> (id, frequency) pairs flattened into a list of size 2*<n>
# field is one of CQI_CONST_FIELD_MATCH, CQI_CONST_FIELD_TARGET,
# CQI_CONST_FIELD_KEYWORD
# NB: pairs are sorted by frequency desc.
""" NOTE: frequency distribution of pairs of tokens """
CQP_FDIST_2 = 0x1511
# INPUT: (STRING subcorpus, INT cutoff, BYTE field1, STRING attribute1,
#         BYTE field2, STRING attribute2)
# OUTPUT: CQI_DATA_INT_LIST
# returns <n> (id1, id2, frequency) pairs flattened into a list of size 3*<n>
# NB: triples are sorted by frequency desc.


""" 4. Constant Definitions """
CONST_FALSE = 0x00
CONST_NO = 0x00
CONST_TRUE = 0x01
CONST_YES = 0x01
"""
" NOTE: The following constants specify which field will be returned by
"       CQI_CQP_DUMP_SUBCORPUS and some other subcorpus commands.
"""
CONST_FIELD_MATCH = 0x10
CONST_FIELD_MATCHEND = 0x11
"""
" NOTE: The constants specifiying target0 .. target9 are guaranteed to have the
" numerical values 0 .. 9, so clients do not need to look up the constant
" values if they're handling arbitrary targets.
"""
CONST_FIELD_TARGET_0 = 0x00
CONST_FIELD_TARGET_1 = 0x01
CONST_FIELD_TARGET_2 = 0x02
CONST_FIELD_TARGET_3 = 0x03
CONST_FIELD_TARGET_4 = 0x04
CONST_FIELD_TARGET_5 = 0x05
CONST_FIELD_TARGET_6 = 0x06
CONST_FIELD_TARGET_7 = 0x07
CONST_FIELD_TARGET_8 = 0x08
CONST_FIELD_TARGET_9 = 0x09
"""
" NOTE: The following constants are provided for backward compatibility with
"       traditional CQP field names & while the generalised target concept
"       isn't yet implemented in the CQPserver.
"""
CONST_FIELD_TARGET = 0x00
CONST_FIELD_KEYWORD = 0x09
""" NOTE: CQi version is CQI_MAJOR_VERSION.CQI_MINOR_VERSION """
MAJOR_VERSION = 0x00
MINOR_VERSION = 0x01


""" 5. CQi lookup dictionary. """
lookup = {
    257: 'CQI_STATUS_OK',
    258: 'CQI_STATUS_CONNECT_OK',
    259: 'CQI_STATUS_BYE_OK',
    260: 'CQI_STATUS_PING_OK',
    513: 'CQI_ERROR_GENERAL_ERROR',
    514: 'CQI_ERROR_CONNECT_REFUSED',
    515: 'CQI_ERROR_USER_ABORT',
    516: 'CQI_ERROR_SYNTAX_ERROR',
    769: 'CQI_DATA_BYTE',
    770: 'CQI_DATA_BOOL',
    771: 'CQI_DATA_INT',
    772: 'CQI_DATA_STRING',
    773: 'CQI_DATA_BYTE_LIST',
    774: 'CQI_DATA_BOOL_LIST',
    775: 'CQI_DATA_INT_LIST',
    776: 'CQI_DATA_STRING_LIST',
    777: 'CQI_DATA_INT_INT',
    778: 'CQI_DATA_INT_INT_INT_INT',
    779: 'CQI_DATA_INT_TABLE',
    1025: 'CQI_CL_ERROR_NO_SUCH_ATTRIBUTE',
    1026: 'CQI_CL_ERROR_WRONG_ATTRIBUTE_TYPE',
    1027: 'CQI_CL_ERROR_OUT_OF_RANGE',
    1028: 'CQI_CL_ERROR_REGEX',
    1029: 'CQI_CL_ERROR_CORPUS_ACCESS',
    1030: 'CQI_CL_ERROR_OUT_OF_MEMORY',
    1031: 'CQI_CL_ERROR_INTERNAL',
    1281: 'CQI_CQP_ERROR_GENERAL',
    1282: 'CQI_CQP_ERROR_NO_SUCH_CORPUS',
    1283: 'CQI_CQP_ERROR_INVALID_FIELD',
    1284: 'CQI_CQP_ERROR_OUT_OF_RANGE',
    4353: 'CQI_CTRL_CONNECT',
    4354: 'CQI_CTRL_BYE',
    4355: 'CQI_CTRL_USER_ABORT',
    4356: 'CQI_CTRL_PING',
    4357: 'CQI_CTRL_LAST_GENERAL_ERROR',
    4609: 'CQI_ASK_FEATURE_CQI_1_0',
    4610: 'CQI_ASK_FEATURE_CL_2_3',
    4611: 'CQI_ASK_FEATURE_CQP_2_3',
    4865: 'CQI_CORPUS_LIST_CORPORA',
    4867: 'CQI_CORPUS_CHARSET',
    4868: 'CQI_CORPUS_PROPERTIES',
    4869: 'CQI_CORPUS_POSITIONAL_ATTRIBUTES',
    4870: 'CQI_CORPUS_STRUCTURAL_ATTRIBUTES',
    4871: 'CQI_CORPUS_STRUCTURAL_ATTRIBUTE_HAS_VALUES',
    4872: 'CQI_CORPUS_ALIGNMENT_ATTRIBUTES',
    4873: 'CQI_CORPUS_FULL_NAME',
    4874: 'CQI_CORPUS_INFO',
    4875: 'CQI_CORPUS_DROP_CORPUS',
    5121: 'CQI_CL_ATTRIBUTE_SIZE',
    5122: 'CQI_CL_LEXICON_SIZE',
    5123: 'CQI_CL_DROP_ATTRIBUTE',
    5124: 'CQI_CL_STR2ID',
    5125: 'CQI_CL_ID2STR',
    5126: 'CQI_CL_ID2FREQ',
    5127: 'CQI_CL_CPOS2ID',
    5128: 'CQI_CL_CPOS2STR',
    5129: 'CQI_CL_CPOS2STRUC',
    5130: 'CQI_CL_CPOS2ALG',
    5131: 'CQI_CL_STRUC2STR',
    5132: 'CQI_CL_ID2CPOS',
    5133: 'CQI_CL_IDLIST2CPOS',
    5134: 'CQI_CL_REGEX2ID',
    5135: 'CQI_CL_STRUC2CPOS',
    5136: 'CQI_CL_ALG2CPOS',
    5152: 'CQI_CL_CPOS2LBOUND',
    5153: 'CQI_CL_CPOS2RBOUND',
    5377: 'CQI_CQP_QUERY',
    5378: 'CQI_CQP_LIST_SUBCORPORA',
    5379: 'CQI_CQP_SUBCORPUS_SIZE',
    5380: 'CQI_CQP_SUBCORPUS_HAS_FIELD',
    5381: 'CQI_CQP_DUMP_SUBCORPUS',
    5385: 'CQI_CQP_DROP_SUBCORPUS',
    5392: 'CQI_CQP_FDIST_1',
    5393: 'CQI_CQP_FDIST_2'
}



from typing import Dict, Optional, Type
from .api import specification


class CQiException(Exception):
    '''
    A base class from which all other exceptions inherit.
    If you want to catch all errors that the CQi package might raise,
    catch this base exception.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: Optional[int] = None
        self.description: Optional[str] = None


class Error(CQiException):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.ERROR


class ErrorGeneralError(Error):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.ERROR_GENERAL_ERROR


class ErrorConnectRefused(Error):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.ERROR_CONNECT_REFUSED


class ErrorUserAbort(Error):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.ERROR_USER_ABORT


class ErrorSyntaxError(Error):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.ERROR_SYNTAX_ERROR


class CLError(CQiException):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CL_ERROR


class CLErrorNoSuchAttribute(CLError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CL_ERROR_NO_SUCH_ATTRIBUTE
        self.description: str = "CQi server couldn't open attribute"


class CLErrorWrongAttributeType(CLError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CL_ERROR_WRONG_ATTRIBUTE_TYPE


class CLErrorOutOfRange(CLError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CL_ERROR_OUT_OF_RANGE


class CLErrorRegex(CLError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CL_ERROR_REGEX


class CLErrorCorpusAccess(CLError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CL_ERROR_CORPUS_ACCESS


class CLErrorOutOfMemory(CLError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CL_ERROR_OUT_OF_MEMORY
        self.description: str = (
            'CQi server has run out of memory; try discarding some other '
            'corpora and/or subcorpora'
        )


class CLErrorInternal(CLError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CL_ERROR_INTERNAL
        self.description: str = (
            "The classical 'please contact technical support' error"
        )


class CQPError(CQiException):
    # CQP error messages yet to be defined
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CQP_ERROR


class CQPErrorGeneral(CQPError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CQP_ERROR_GENERAL


class CQPErrorNoSuchCorpus(CQPError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CQP_ERROR_NO_SUCH_CORPUS


class CQPErrorInvalidField(CQPError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CQP_ERROR_INVALID_FIELD


class CQPErrorOutOfRange(CQPError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code: int = specification.CQP_ERROR_OUT_OF_RANGE
        self.description: str = 'A number is out of range'


lookup: Dict[int, Type[CQiException]] = {
    specification.ERROR: Error,
    specification.ERROR_GENERAL_ERROR: ErrorGeneralError,
    specification.ERROR_CONNECT_REFUSED: ErrorConnectRefused,
    specification.ERROR_USER_ABORT: ErrorUserAbort,
    specification.ERROR_SYNTAX_ERROR: ErrorSyntaxError,
    specification.CL_ERROR: CLError,
    specification.CL_ERROR_NO_SUCH_ATTRIBUTE: CLErrorNoSuchAttribute,
    specification.CL_ERROR_WRONG_ATTRIBUTE_TYPE: CLErrorWrongAttributeType,
    specification.CL_ERROR_OUT_OF_RANGE: CLErrorOutOfRange,
    specification.CL_ERROR_REGEX: CLErrorRegex,
    specification.CL_ERROR_CORPUS_ACCESS: CLErrorCorpusAccess,
    specification.CL_ERROR_OUT_OF_MEMORY: CLErrorOutOfMemory,
    specification.CL_ERROR_INTERNAL: CLErrorInternal,
    specification.CQP_ERROR: CQPError,
    specification.CQP_ERROR_GENERAL: CQPErrorGeneral,
    specification.CQP_ERROR_NO_SUCH_CORPUS: CQPErrorNoSuchCorpus,
    specification.CQP_ERROR_INVALID_FIELD: CQPErrorInvalidField,
    specification.CQP_ERROR_OUT_OF_RANGE: CQPErrorOutOfRange
}



from typing import Dict, List, Type, TYPE_CHECKING
if TYPE_CHECKING:
    from ..client import CQiClient


class Model:
    '''
    A base class for representing a single object on the server.
    '''

    def __init__(
        self,
        attrs: Dict = None,
        client: 'CQiClient' = None,
        collection: 'Collection' = None
    ):
        #: A client pointing at the server that this object is on.
        self.client: 'CQiClient' = client

        #: The collection that this model is part of.
        self.collection: Collection = collection

        #: The raw representation of this object from the API
        self.attrs: Dict = attrs or {}

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.api_name}>'

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, self.__class__)
            and self.api_name == other.api_name
        )

    def __hash__(self) -> int:
        return hash(f'{self.__class__.__name__}:{self.api_name}')

    @property
    def api_name(self) -> str:
        raise NotImplementedError

    def reload(self):
        '''
        Load this object from the server again and update ``attrs`` with the
        new data.
        '''
        self.attrs = self.collection.get(self.api_name).attrs


class Collection:
    '''
    A base class for representing all objects of a particular type on the
    server.
    '''

    #: The type of object this collection represents, set by subclasses
    model: Type[Model] = Model

    def __init__(self, client: 'CQiClient' = None):
        #: The client pointing at the server that this collection of objects
        #: is on.
        self.client: 'CQiClient' = client

    def list(self) -> List[Model]:
        raise NotImplementedError

    def get(self) -> Model:
        raise NotImplementedError

    def prepare_model(self, attrs) -> Model:
        '''
        Create a model from a set of attributes.
        '''
        if isinstance(attrs, Model):
            attrs.client = self.client
            attrs.collection = self
            return attrs
        elif isinstance(attrs, dict):
            return self.model(
                attrs=attrs,
                client=self.client,
                collection=self
            )
        else:
            raise Exception(f"Can't create {self.model.__name__} from {attrs}")



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

version = None
exec(open('cqi/version.py').read())

setuptools.setup(
    name='cqi',
    version=version,
    author='Patrick Jentsch',
    author_email='patrickjentsch@gmx.net',
    description=(
        'A Python library for the IMS Open Corpus Workbench (CWB) '
        'corpus query interface (CQi) API.'
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Pevtrick/cqi-py',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.5',
)



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .status import StatusByeOk, StatusConnectOk, StatusPingOk
from .api import APIClient
from .models.corpora import CorpusCollection



class CQiClient:
    '''
    A client for communicating with a CQi server.

    Example:
    >>> import cqi
    >>> client = cqi.CQiClient('127.0.0.1')
    >>> client.connect()
    258  # CQI_STATUS_CONNECT_OK
    >>> client.ping()
    260  # CQI_STATUS_PING_OK
    >>> client.disconnect()
    259  # CQI_STATUS_BYE_OK

    Attributes:
    api (APIClient): An API client pointing to the specified CQP server.
    '''

    def __init__(
        self,
        host: str,
        port: int = 4877,
        timeout: float = 60.0,
        version: str = '0.1'
    ):
        '''
        CQiClient constructor

        Args:
        host (str): URL to the CQP server. For example,
            ``cqpserver.localhost`` or ``127.0.0.1``.
        port (int): Port the CQP server listens on. Default: ``4877``
        timeout (float): Default timeout for API calls, in seconds. Default: ``60.0``
        version (str): The version of the CQi protocol to use. Default: ``0.1``
        '''
        self.api: APIClient = APIClient(
            host,
            port=port,
            timeout=timeout,
            version=version
        )

    @property
    def corpora(self) -> CorpusCollection:
        return CorpusCollection(client=self)

    def bye(self) -> 'StatusByeOk':
        return self.api.ctrl_bye()

    def connect(self, username: str, password: str) -> 'StatusConnectOk':
        return self.api.ctrl_connect(username, password)

    def ping(self) -> 'StatusPingOk':
        return self.api.ctrl_ping()

    def user_abort(self):
        self.api.ctrl_user_abort()

    # Method aliases
    disconnect = bye



from typing import Dict, List, Type, TYPE_CHECKING
if TYPE_CHECKING:
    from ..client import CQiClient
    from ..status import StatusOk
    from .attributes import PositionalAttribute
    from .corpora import Corpus
from ..api.specification import (
    CONST_FIELD_KEYWORD,
    CONST_FIELD_MATCH,
    CONST_FIELD_MATCHEND,
    CONST_FIELD_TARGET
)
from .resource import Collection, Model


class Subcorpus(Model):
    @property
    def api_name(self) -> str:
        return self.attrs['api_name']

    @property
    def fields(self) -> Dict[str, int]:
        return self.attrs['fields']

    @property
    def name(self) -> str:
        return self.attrs['name']

    @property
    def size(self) -> int:
        return self.attrs['size']

    def drop(self) -> 'StatusOk':
        ''' delete a subcorpus from memory '''
        return self.client.api.cqp_drop_subcorpus(self.api_name)

    def dump(self, field: int, first: int, last: int) -> List[int]:
        '''
        Dump the values of <field> for match ranges <first> .. <last> in
        subcorpus. <field> is one of the CQI_CONST_FIELD_* constants.
        '''
        return self.client.api.cqp_dump_subcorpus(
            self.api_name,
            field,
            first,
            last
        )

    def fdist_1(
        self,
        cutoff: int,
        field: int,
        attribute: 'PositionalAttribute'
    ) -> List[int]:
        '''
        frequency distribution of single tokens

        returns <n> (id, frequency) pairs flattened into a list of size 2*<n>

        field is one of
        - CQI_CONST_FIELD_MATCH
        - CQI_CONST_FIELD_TARGET
        - CQI_CONST_FIELD_KEYWORD

        NB: pairs are sorted by frequency desc.
        '''
        return self.client.api.cqp_fdist_1(
            self.api_name,
            cutoff,
            field,
            attribute.api_name
        )

    def fdist_2(
        self,
        cutoff: int,
        field_1: int,
        attribute_1: 'PositionalAttribute',
        field_2: int,
        attribute_2: 'PositionalAttribute'
    ) -> List[int]:
        '''
        frequency distribution of pairs of tokens

        returns <n> (id1, id2, frequency) pairs flattened into a list of size
        3*<n>

        NB: triples are sorted by frequency desc.
        '''
        return self.client.api.cqp_fdist_2(
            self.api_name,
            cutoff,
            field_1,
            attribute_1.api_name,
            field_2,
            attribute_2.api_name
        )


class SubcorpusCollection(Collection):
    model: Type[Subcorpus] = Subcorpus

    def __init__(self, client: 'CQiClient' = None, corpus: 'Corpus' = None):
        super().__init__(client=client)
        self.corpus: 'Corpus' = corpus

    def _get(self, subcorpus_name: str) -> Dict:
        api_name: str = f'{self.corpus.api_name}:{subcorpus_name}'
        fields: Dict[str, int] = {}
        if self.client.api.cqp_subcorpus_has_field(api_name, CONST_FIELD_MATCH):
            fields['match'] = CONST_FIELD_MATCH
        if self.client.api.cqp_subcorpus_has_field(api_name, CONST_FIELD_MATCHEND):
            fields['matchend'] = CONST_FIELD_MATCHEND
        if self.client.api.cqp_subcorpus_has_field(api_name, CONST_FIELD_TARGET):
            fields['target'] = CONST_FIELD_TARGET
        if self.client.api.cqp_subcorpus_has_field(api_name, CONST_FIELD_KEYWORD):
            fields['keyword'] = CONST_FIELD_KEYWORD
        return {
            'api_name': api_name,
            'fields': fields,
            'name': subcorpus_name,
            'size': self.client.api.cqp_subcorpus_size(api_name)
        }

    def get(self, subcorpus_name: str) -> Subcorpus:
        return self.prepare_model(self._get(subcorpus_name))

    def list(self) -> List[Subcorpus]:
        return [
            self.get(x) for x in
            self.client.api.cqp_list_subcorpora(self.corpus.api_name)
        ]
